"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""
from dataclasses import dataclass

from sc_client.constants import sc_type

from sc_kpm.sc_keynodes import Idtf, ScKeynodes


@dataclass(frozen=True)
class CommonIdentifiers:
    ACTION: Idtf = "action"
    EXACT_VALUE: Idtf = "exact_value"
    RREL_DYNAMIC_ARGUMENT: Idtf = "rrel_dynamic_argument"
    RREL_ONE: Idtf = "rrel_1"
    RREL_TWO: Idtf = "rrel_2"
    RREL_LAST: Idtf = "rrel_last"
    NREL_BASIC_SEQUENCE: Idtf = "nrel_basic_sequence"
    NREL_SYSTEM_IDENTIFIER: Idtf = "nrel_system_identifier"
    NREL_RESULT: Idtf = "nrel_result"
    CONCEPT_FILENAME: Idtf = "concept_filename"


@dataclass(frozen=True)
class ActionStatus:
    ACTION_INITIATED: Idtf = "action_initiated"
    ACTION_FINISHED: Idtf = "action_finished"
    ACTION_FINISHED_SUCCESSFULLY: Idtf = "action_finished_successfully"
    ACTION_FINISHED_UNSUCCESSFULLY: Idtf = "action_finished_unsuccessfully"


@dataclass(frozen=True)
class ScAlias:
    ACTION_NODE: str = "_action_node"
    RELATION_ARC: str = "_relation_arc"
    MEMBERSHIP_ARC: str = "_membership_arc"
    ELEMENT: str = "_element"
    LINK: str = "_link"


class _IdentifiersResolver:
    """
    Class for resolving common identifiers and action status identifiers.
    It confirms the presence of all of them in the KB.
    """

    is_resolved = False

    @classmethod
    def resolve(cls) -> None:
        if cls.is_resolved:
            return
        types_map = {
            CommonIdentifiers.ACTION: sc_type.CONST_NODE_CLASS,
            CommonIdentifiers.EXACT_VALUE: sc_type.CONST_NODE_CLASS,
            CommonIdentifiers.RREL_DYNAMIC_ARGUMENT: sc_type.CONST_NODE_ROLE,
            CommonIdentifiers.RREL_ONE: sc_type.CONST_NODE_ROLE,
            CommonIdentifiers.RREL_TWO: sc_type.CONST_NODE_ROLE,
            CommonIdentifiers.RREL_LAST: sc_type.CONST_NODE_ROLE,
            CommonIdentifiers.NREL_BASIC_SEQUENCE: sc_type.CONST_NODE_NON_ROLE,
            CommonIdentifiers.NREL_SYSTEM_IDENTIFIER: sc_type.CONST_NODE_NON_ROLE,
            CommonIdentifiers.NREL_RESULT: sc_type.CONST_NODE_NON_ROLE,
            CommonIdentifiers.CONCEPT_FILENAME: sc_type.CONST_NODE_CLASS,
            ActionStatus.ACTION_INITIATED: sc_type.CONST_NODE_CLASS,
            ActionStatus.ACTION_FINISHED: sc_type.CONST_NODE_CLASS,
            ActionStatus.ACTION_FINISHED_SUCCESSFULLY: sc_type.CONST_NODE_CLASS,
            ActionStatus.ACTION_FINISHED_UNSUCCESSFULLY: sc_type.CONST_NODE_CLASS,
        }

        for idtf, type_value in types_map.items():
            ScKeynodes.resolve(idtf, type_value)

        cls.is_resolved = True



"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""

import warnings
from logging import Logger, getLogger
from typing import Dict, Optional

from sc_client import client
from sc_client.client import erase_elements
from sc_client.constants.exceptions import InvalidValueError
from sc_client.constants.sc_type import CONST_NODE_ROLE, ScType
from sc_client.models import ScAddr, ScIdtfResolveParams

Idtf = str


class ScKeynodesMeta(type):
    """Metaclass to use ScKeynodes without creating an instance of a class"""

    def __init__(cls, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        cls._dict: Dict[Idtf, ScAddr] = {}
        cls._logger: Logger = getLogger(f"{__name__}.{cls.__name__}")
        cls._min_rrel_index: int = 1
        cls._max_rrel_index: int = 10

    def __call__(cls, *args, **kwargs) -> None:
        raise TypeError(f"Use {cls.__name__} without initialization")

    def __getitem__(cls, identifier: Idtf) -> ScAddr:
        """Get keynode, cannot be invalid ScAddr(0)"""
        addr = cls.get(identifier)  # pylint: disable=no-value-for-parameter
        if not addr.is_valid():
            cls._logger.error("Failed to get ScAddr by %s keynode: ScAddr is invalid", identifier)
            raise InvalidValueError(f"ScAddr of {identifier} is invalid")
        return addr

    def erase(cls, identifier: Idtf) -> bool:
        """Erase keynode from the kb and memory and return boolean status"""
        addr = cls.__getitem__(identifier)  # pylint: disable=no-value-for-parameter
        del cls._dict[identifier]
        return erase_elements(addr)

    def delete(cls, identifier: Idtf) -> bool:
        warnings.warn(
            "ScKeynodesMeta 'delete' method is deprecated. Use `erase` method instead.",
            DeprecationWarning,
        )
        return cls.erase(identifier)

    def get(cls, identifier: Idtf) -> ScAddr:
        """Get keynode, can be ScAddr(0)"""
        return cls.resolve(identifier, None)  # pylint: disable=no-value-for-parameter

    def resolve(cls, identifier: Idtf, sc_type: Optional[ScType]) -> ScAddr:
        """Get keynode. If sc_type is valid, an element will be created in the KB"""
        addr = cls._dict.get(identifier)
        if addr is None:
            params = ScIdtfResolveParams(idtf=identifier, type=sc_type)
            addr = client.resolve_keynodes(params)[0]
            if addr.is_valid():
                cls._dict[identifier] = addr
            cls._logger.debug(
                "Resolved %s identifier with type %s: %s",
                repr(identifier),
                repr(sc_type),
                repr(addr),
            )
        return addr

    def rrel_index(cls, index: int) -> ScAddr:
        """Get rrel_i node. Max rrel index is 10. Min rrel is 1."""
        if not isinstance(index, int):
            raise TypeError("Index of rrel node must be int")
        if index > cls._max_rrel_index:
            raise KeyError(f"You cannot use rrel more than {cls._max_rrel_index}")
        if index < cls._min_rrel_index:
            raise KeyError(f"You cannot use rrel less than {cls._min_rrel_index}")
        return cls.resolve(f"rrel_{index}", CONST_NODE_ROLE)  # pylint: disable=no-value-for-parameter


class ScKeynodes(metaclass=ScKeynodesMeta):
    """Class which provides the ability to cache the identifier and ScAddr of keynodes stored in the KB."""



"""
This source file is part of an OSTIS project. For the latest info, see https:#github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https:#opensource.org/licenses/MIT)
"""

from __future__ import annotations

import signal
from abc import ABC, abstractmethod
from logging import Logger, getLogger
from typing import Callable

from sc_client import client

from sc_kpm.identifiers import _IdentifiersResolver
from sc_kpm.sc_module import ScModuleAbstract


class ScServerAbstract(ABC):
    """ScServer connects to server and stores"""

    @abstractmethod
    def connect(self) -> _Finisher:
        """Connect to server"""

    @abstractmethod
    def disconnect(self) -> None:
        """Disconnect from the server"""

    @abstractmethod
    def add_modules(self, *modules: ScModuleAbstract) -> None:
        """Add modules to the server and register them if server is registered"""

    @abstractmethod
    def remove_modules(self, *modules: ScModuleAbstract) -> None:
        """Remove modules from the server and unregister them if server is registered"""

    @abstractmethod
    def clear_modules(self) -> None:
        """Remove all modules from the server and unregister them if server is registered"""

    def register_modules(self) -> _Finisher:
        """Register all modules in the server"""

    def unregister_modules(self) -> None:
        """Unregister all modules from the server"""

    def start(self) -> _Finisher:
        """Connect and register modules"""

    def stop(self) -> None:
        """Disconnect and unregister modules"""


class ScServer(ScServerAbstract):
    def __init__(self, sc_server_url: str) -> None:
        self._url: str = sc_server_url
        self._modules: set[ScModuleAbstract] = set()
        self.is_registered = False
        self.logger = getLogger(f"{self.__module__}.{self.__class__.__name__}")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(map(repr, self._modules))})"

    def connect(self) -> _Finisher:
        client.connect(self._url)
        self.logger.info("Connected by url: %s", repr(self._url))
        _IdentifiersResolver.resolve()
        return _Finisher(self.disconnect, self.logger)

    def disconnect(self) -> None:
        client.disconnect()
        self.logger.info("Disconnected from url: %s", repr(self._url))

    def add_modules(self, *modules: ScModuleAbstract) -> None:
        if self.is_registered:
            self._register(*modules)
        self._modules |= {*modules}
        self.logger.info("Added modules: %s", ", ".join(map(repr, modules)))

    def remove_modules(self, *modules: ScModuleAbstract) -> None:
        if self.is_registered:
            self._unregister(*modules)
        self._modules -= {*modules}
        self.logger.info("Removed modules: %s", ", ".join(map(repr, modules)))

    def clear_modules(self) -> None:
        if self.is_registered:
            self._unregister(*self._modules)
        self.logger.info("Removed all modules: %s", ", ".join(map(repr, self._modules)))
        self._modules.clear()

    def register_modules(self) -> _Finisher:
        if self.is_registered:
            self.logger.warning("Modules are already registered")
        else:
            self._register(*self._modules)
            self.is_registered = True
            self.logger.info("Registered modules successfully")
        return _Finisher(self.unregister_modules, self.logger)

    def unregister_modules(self) -> None:
        if not self.is_registered:
            self.logger.warning("Modules are already unregistered")
            return
        self._unregister(*self._modules)
        self.is_registered = False
        self.logger.info("Unregistered modules successfully")

    def start(self) -> _Finisher:
        self.connect()
        self.register_modules()
        return _Finisher(self.stop, self.logger)

    def stop(self) -> None:
        self.unregister_modules()
        self.disconnect()

    def _register(self, *modules: ScModuleAbstract) -> None:
        if not client.is_connected():
            self.logger.error("Failed to register: connection lost")
            raise ConnectionError(f"Connection to url {repr(self._url)} lost")
        for module in modules:
            if not isinstance(module, ScModuleAbstract):
                self.logger.error("Failed to register: type of %s is not ScModule", repr(module))
                raise TypeError(f"{repr(module)} is not ScModule")
            module._register()  # pylint: disable=protected-access

    def _unregister(self, *modules: ScModuleAbstract) -> None:
        if not client.is_connected():
            self.logger.error("Failed to unregister: connection to %s lost", repr(self._url))
            raise ConnectionError(f"Connection to {repr(self._url)} lost")
        for module in modules:
            module._unregister()  # pylint: disable=protected-access

    def serve(self) -> None:
        """Serve agents until a SIGINT signal (^C, or stop in IDE) is received"""

        signal.signal(signal.SIGINT, lambda *_: self.logger.info("^C SIGINT was interrupted"))
        signal.pause()


class _Finisher:
    """Class for calling finish method in with-statement"""

    def __init__(self, finish_method: Callable[[], None], logger: Logger) -> None:
        self._finish_method = finish_method
        self._logger = logger

    def __enter__(self) -> None:
        pass  # Interaction through the beginning method (with server.start_method(): ...)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if exc_val is not None:
            self._logger.error("Raised error %s, finishing", repr(exc_val))
        self._finish_method()



from typing import Iterable, Iterator

from sc_client.client import get_link_content
from sc_client.models import ScAddr
from sc_client.models.sc_construction import ScLinkContent, ScLinkContentData


def iter_link_contents_data(
    contents: Iterable[ScLinkContent],
) -> Iterator[ScLinkContentData]:
    """Iterate by data in link contents"""
    for content in contents:
        yield content.data


def iter_links_data(links: Iterable[ScAddr]) -> Iterator[ScLinkContentData]:
    """Iterate by contents data in links"""
    contents = get_link_content(*links)
    return iter_link_contents_data(contents)



from typing import Iterator, List

from sc_client.client import generate_by_template, search_by_template
from sc_client.constants import sc_type
from sc_client.models import ScAddr, ScTemplate

from sc_kpm.sc_keynodes import ScKeynodes
from sc_kpm.sc_sets.sc_set import ScSet


class ScNumberedSet(ScSet):
    """
    ScNumberedSet is a class for handling numbered set structure in kb.

    It has main set_node and arc elements:
    Arcs to each element marked with 'rrel_1', 'rrel_2', and so on nodes.
    """

    def add(self, *elements: ScAddr) -> None:
        """Add elements to ScNumberedSet"""
        if elements:
            template = ScTemplate()
            for index, element in enumerate(elements, len(self) + 1):
                template.quintuple(
                    self._set_node,
                    sc_type.VAR_PERM_POS_ARC,
                    element,
                    sc_type.VAR_PERM_POS_ARC,
                    ScKeynodes.rrel_index(index),
                )
            generate_by_template(template)

    def __iter__(self) -> Iterator[ScAddr]:
        return iter(self.elements_list)

    @property
    def elements_list(self) -> List[ScAddr]:
        """Iterate by ScNumberedSet elements"""
        templ = ScTemplate()
        templ.quintuple(
            self._set_node,
            sc_type.VAR_PERM_POS_ARC,
            sc_type.UNKNOWN,
            sc_type.VAR_PERM_POS_ARC,
            sc_type.VAR_NODE_ROLE,
        )
        results = search_by_template(templ)
        sorted_results = sorted((result for result in results), key=lambda res: res[4].value)
        # Sort rrel elements addrs
        return [result[2] for result in sorted_results]

    def __getitem__(self, i: int) -> ScAddr:
        templ = ScTemplate()
        templ.quintuple(
            self._set_node,
            sc_type.VAR_PERM_POS_ARC,
            sc_type.UNKNOWN,
            sc_type.VAR_PERM_POS_ARC,
            ScKeynodes.rrel_index(i + 1),
        )
        results = search_by_template(templ)
        if not results:
            raise KeyError("No element by index")
        return results[0][2]

    def remove(self, *elements: ScAddr) -> None:
        """Clear and add existing elements without given ones"""
        # TODO: optimize
        elements_new = [element for element in self.elements_list if element not in elements]
        self.clear()
        self.add(*elements_new)



"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""

from abc import ABC, abstractmethod
from logging import getLogger
from typing import Optional, Union

from sc_client import client
from sc_client.constants import sc_type
from sc_client.constants.common import ScEventType
from sc_client.constants.exceptions import InvalidValueError
from sc_client.models import ScAddr, ScEventSubscription, ScEventSubscriptionParams

from sc_kpm.identifiers import ActionStatus
from sc_kpm.sc_keynodes import Idtf, ScKeynodes
from sc_kpm.sc_result import ScResult
from sc_kpm.utils.action_utils import check_action_class


class ScAgentAbstract(ABC):
    def __init__(self, event_element: ScAddr, event_type: ScEventType) -> None:
        self._event_element = event_element
        self._event_type = event_type
        self._event: Optional[ScEventSubscription] = None
        self.logger = getLogger(f"{self.__module__}.{self.__class__.__name__}")

    @abstractmethod
    def __repr__(self) -> str:
        pass

    def _register(self) -> None:
        if self._event is not None:
            self.logger.warning("Almost registered")
            return
        event_params = ScEventSubscriptionParams(self._event_element, self._event_type, self._callback)
        self._event = client.create_elementary_event_subscriptions(event_params)[0]
        self.logger.info(
            "Registered with ScEvent: %s - %s",
            repr(self._event_element),
            repr(self._event_type),
        )

    def _unregister(self) -> None:
        if self._event is None:
            self.logger.warning("ScEvent was already destroyed or not registered")
            return
        client.destroy_elementary_event_subscriptions(self._event)
        self._event = None
        self.logger.info(
            "Unregistered ScEvent: %s - %s",
            repr(self._event_element),
            repr(self._event_type),
        )

    def _callback(self, event_element: ScAddr, event_connector: ScAddr, action_element: ScAddr) -> ScResult:
        return self.on_event(event_element, event_connector, action_element)

    @abstractmethod
    def on_event(self, event_element: ScAddr, event_connector: ScAddr, action_element: ScAddr) -> ScResult:
        pass


class ScAgent(ScAgentAbstract, ABC):
    def __init__(self, event_element: Union[Idtf, ScAddr], event_type: ScEventType) -> None:
        if isinstance(event_element, Idtf):
            event_element = ScKeynodes.resolve(event_element, sc_type.CONST_NODE_CLASS)
        if not event_element.is_valid():
            self.logger.error("Failed to initialize ScAgent: event_class is invalid")
            raise InvalidValueError(f"event_class of {self.__class__.__name__} is invalid")
        super().__init__(event_element, event_type)

    def __repr__(self) -> str:
        return f"ScAgent(event_class='{self._event_element}', event_type={repr(self._event_type)})"


class ScAgentClassic(ScAgent, ABC):
    def __init__(
        self,
        action_class_name: Idtf,
        event_element: Union[Idtf, ScAddr] = ActionStatus.ACTION_INITIATED,
        event_type: ScEventType = ScEventType.AFTER_GENERATE_OUTGOING_ARC,
    ) -> None:
        super().__init__(event_element, event_type)
        self._action_class_name = action_class_name
        self._action_class = ScKeynodes.resolve(action_class_name, sc_type.CONST_NODE_CLASS)

    def __repr__(self) -> str:
        description = f"ClassicScAgent(action_class_name={repr(self._action_class_name)}"
        if self._event_element != ScKeynodes.get(ActionStatus.ACTION_INITIATED):
            description = f"{description}, event_class={repr(self._event_element)}"
        if self._event_type != ScEventType.AFTER_GENERATE_OUTGOING_ARC:
            description = f"{description}, event_type={repr(self._event_type)}"
        return description + ")"

    def _callback(self, event_element: ScAddr, event_connector: ScAddr, action_element: ScAddr) -> ScResult:
        if not check_action_class(self._action_class, action_element):
            return ScResult.SKIP
        self.logger.info("Confirmed action class")
        return self.on_event(event_element, event_connector, action_element)



"""
This source file is part of an OSTIS project. For the latest info, see https:#github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https:#opensource.org/licenses/MIT)
"""

from enum import IntEnum, auto


class ScResult(IntEnum):
    ERROR = 0  # unknown error, should be 0 anytime
    OK = 1  # no errors, should be 1 anytime
    ERROR_INVALID_PARAMS = auto()  # invalid function parameters error
    ERROR_INVALID_TYPE = auto()  # invalid type error
    ERROR_IO = auto()  # input/output error
    ERROR_INVALID_STATE = auto()  # invalid state of processed object
    ERROR_NOT_FOUND = auto()  # item not found
    ERROR_NO_WRITE_RIGHTS = auto()  # no rights to change or erase object
    ERROR_NO_READ_RIGHTS = auto()  # no rights to read object
    NO = auto()  # no result
    UNKNOWN = auto()  # result unknown
    SKIP = auto()  # skip agent



"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""

import logging


def set_root_config(root_name: str) -> None:
    logging.getLogger(root_name).addHandler(logging.NullHandler())



"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""

from sc_kpm import utils
from sc_kpm.logging import set_root_config
from sc_kpm.sc_agent import ScAgent, ScAgentClassic
from sc_kpm.sc_keynodes import ScKeynodes
from sc_kpm.sc_module import ScModule
from sc_kpm.sc_result import ScResult
from sc_kpm.sc_server import ScServer

set_root_config(__name__)



from common_tests import BaseTestCase
from sc_client import client
from sc_client.client import erase_elements, get_elements_types
from sc_client.constants import sc_type
from sc_client.constants.exceptions import InvalidValueError
from sc_client.models import ScAddr, ScIdtfResolveParams

from sc_kpm import ScKeynodes


class KeynodesTests(BaseTestCase):
    def test_get_existed_keynode(self):
        idtf = "idtf_existed_keynode"
        params = ScIdtfResolveParams(idtf=idtf, type=sc_type.CONST_NODE)
        addr = client.resolve_keynodes(params)[0]
        result = ScKeynodes[idtf]
        self.assertEqual(result, addr)

    def test_get_unknown_idtf(self):
        idtf = "idtf_unknown_idtf"
        self.assertRaises(InvalidValueError, ScKeynodes.__getitem__, idtf)
        self.assertEqual(ScKeynodes.get(idtf), ScAddr(0))

    def test_resolve_keynode(self):
        idtf = "idtf_new_keynode"
        addr = ScKeynodes.resolve(idtf, sc_type.CONST_NODE)
        self.assertTrue(erase_elements(addr))
        self.assertTrue(addr.is_valid())

    def test_erase_keynode(self):
        idtf = "idtf_to_erase_keynode"
        ScKeynodes.resolve(idtf, sc_type.CONST_NODE)
        self.assertTrue(ScKeynodes.erase(idtf))
        self.assertFalse(ScKeynodes.get(idtf).is_valid())
        self.assertRaises(InvalidValueError, ScKeynodes.erase, idtf)

    def test_keynodes_initialization(self):
        self.assertRaises(TypeError, ScKeynodes)

    def test_rrel(self):
        rrel_1 = ScKeynodes.rrel_index(1)
        self.assertTrue(rrel_1.is_valid())
        self.assertTrue(get_elements_types(rrel_1)[0].is_role())

    def test_max_rrel(self):
        self.assertRaises(KeyError, ScKeynodes.rrel_index, ScKeynodes._max_rrel_index + 1)

    def test_min_rrel(self):
        self.assertRaises(KeyError, ScKeynodes.rrel_index, ScKeynodes._min_rrel_index - 1)

    def test_wrong_rrel(self):
        self.assertRaises(TypeError, ScKeynodes.rrel_index, "str")



import sys
from pathlib import Path

from setuptools import find_packages, setup

DIRECTORY_PATH = Path(__file__).parent
README = (DIRECTORY_PATH / "README.md").read_text()

VERSION = "0.4.0"
INSTALL_REQUIRES = ["py-sc-client>=0.4.0"]
CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 8)


if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write(
        # pylint: disable=consider-using-f-string
        """
            ==========================
            Unsupported Python version
            ==========================
            This version of py-sc-client requires at least Python {}.{}, but
            you're trying to install it on Python {}.{}. To resolve this,
            consider upgrading to a supported Python version.
            """.format(
            *REQUIRED_PYTHON, *CURRENT_PYTHON
        )
    )
    sys.exit(1)


setup(
    name="py-sc-kpm",
    version=VERSION,
    description="The Python implementation of the knowledge processing module for knowledge manipulations",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/ostis-ai/py-sc-kpm",
    author="ostis-ai",
    license="MIT",
    classifiers=[
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="sc-kpm, kpm, knowledge processing",
    packages=find_packages(where="src", exclude=("tests",)),
    package_dir={"": "src"},
    python_requires=">=3.8, <4",
    install_requires=INSTALL_REQUIRES,
    project_urls={
        "Bug Reports": "https://github.com/ostis-ai/py-sc-kpm/issues",
        "Source": "https://github.com/ostis-ai/py-sc-kpm",
    },
)



"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""

import warnings
from threading import Event
from typing import Dict, List, Tuple, Union

from sc_client import client
from sc_client.client import create_elementary_event_subscriptions, destroy_elementary_event_subscriptions
from sc_client.constants import sc_type
from sc_client.constants.common import ScEventType
from sc_client.models import ScAddr, ScConstruction, ScEventSubscriptionParams, ScTemplate

from sc_kpm.identifiers import ActionStatus, CommonIdentifiers, ScAlias
from sc_kpm.sc_keynodes import Idtf, ScKeynodes
from sc_kpm.sc_result import ScResult
from sc_kpm.sc_sets.sc_structure import ScStructure
from sc_kpm.utils.common_utils import (
    check_connector,
    generate_connector,
    generate_node,
    generate_non_role_relation,
    generate_role_relation,
    search_element_by_role_relation,
)

COMMON_WAIT_TIME: float = 5


def check_action_class(action_class: Union[ScAddr, Idtf], action_node: ScAddr) -> bool:
    action_class = ScKeynodes[action_class] if isinstance(action_class, Idtf) else action_class
    templ = ScTemplate()
    templ.triple(action_class, sc_type.VAR_PERM_POS_ARC, action_node)
    templ.triple(ScKeynodes[CommonIdentifiers.ACTION], sc_type.VAR_PERM_POS_ARC, action_node)
    search_results = client.search_by_template(templ)
    return len(search_results) > 0


def get_action_arguments(action_node: ScAddr, count: int) -> List[ScAddr]:
    arguments = []
    for index in range(1, count + 1):
        argument = search_element_by_role_relation(action_node, ScKeynodes.rrel_index(index))
        arguments.append(argument)
    return arguments


def generate_action_result(action_node: ScAddr, *elements: ScAddr) -> None:
    result_struct_node = ScStructure(*elements).set_node
    generate_non_role_relation(action_node, result_struct_node, ScKeynodes[CommonIdentifiers.NREL_RESULT])


def create_action_result(action_node: ScAddr, *elements: ScAddr) -> None:
    warnings.warn(
        "Action utils 'create_action_result' method is deprecated. Use `generate_action_result` method instead.",
        DeprecationWarning,
    )
    generate_action_result(action_node, *elements)


def get_action_result(action_node: ScAddr) -> ScAddr:
    templ = ScTemplate()
    templ.quintuple(
        action_node,
        sc_type.VAR_COMMON_ARC >> ScAlias.RELATION_ARC,
        sc_type.VAR_NODE_STRUCTURE >> ScAlias.ELEMENT,
        sc_type.VAR_PERM_POS_ARC,
        ScKeynodes[CommonIdentifiers.NREL_RESULT],
    )
    if search_results := client.search_by_template(templ):
        return search_results[0].get(ScAlias.ELEMENT)
    return ScAddr(0)


IsDynamic = bool


def execute_agent(
    arguments: Dict[ScAddr, IsDynamic],
    concepts: List[Idtf],
    initiation: Idtf = ActionStatus.ACTION_INITIATED,
    reaction: Idtf = ActionStatus.ACTION_FINISHED_SUCCESSFULLY,
    wait_time: float = COMMON_WAIT_TIME,
) -> Tuple[ScAddr, bool]:
    action = call_agent(arguments, concepts, initiation)
    wait_agent(wait_time, action)
    result = check_connector(sc_type.VAR_PERM_POS_ARC, ScKeynodes[reaction], action)
    return action, result


def call_agent(
    arguments: Dict[ScAddr, IsDynamic],
    concepts: List[Idtf],
    initiation: Idtf = ActionStatus.ACTION_INITIATED,
) -> ScAddr:
    action = generate_action(*concepts)
    add_action_arguments(action, arguments)
    call_action(action, initiation)
    return action


def generate_action(*concepts: Idtf) -> ScAddr:
    construction = ScConstruction()
    construction.generate_node(sc_type.CONST_NODE, ScAlias.ACTION_NODE)
    for concept in concepts:
        construction.generate_connector(
            sc_type.CONST_PERM_POS_ARC,
            ScKeynodes.resolve(concept, sc_type.CONST_NODE_CLASS),
            ScAlias.ACTION_NODE,
        )
    action_node = client.generate_elements(construction)[0]
    return action_node


def create_action(*concepts: Idtf) -> ScAddr:
    warnings.warn(
        "Action utils 'create_action' method is deprecated. Use `generate_action` method instead.",
        DeprecationWarning,
    )
    return generate_action(*concepts)


def add_action_arguments(action_node: ScAddr, arguments: Dict[ScAddr, IsDynamic]) -> None:
    rrel_dynamic_arg = ScKeynodes[CommonIdentifiers.RREL_DYNAMIC_ARGUMENT]
    argument: ScAddr
    for index, (argument, is_dynamic) in enumerate(arguments.items(), 1):
        if argument.is_valid():
            rrel_i = ScKeynodes.rrel_index(index)
            if is_dynamic:
                dynamic_node = generate_node(sc_type.CONST_NODE)
                generate_role_relation(action_node, dynamic_node, rrel_dynamic_arg, rrel_i)
                generate_connector(sc_type.CONST_TEMP_POS_ARC, dynamic_node, argument)
            else:
                generate_role_relation(action_node, argument, rrel_i)


def execute_action(
    action_node: ScAddr,
    initiation: Idtf = ActionStatus.ACTION_INITIATED,
    reaction: Idtf = ActionStatus.ACTION_FINISHED_SUCCESSFULLY,
    wait_time: float = COMMON_WAIT_TIME,
) -> bool:
    call_action(action_node, initiation)
    wait_agent(wait_time, action_node)
    result = check_connector(sc_type.VAR_PERM_POS_ARC, ScKeynodes[reaction], action_node)
    return result


def call_action(action_node: ScAddr, initiation: Idtf = ActionStatus.ACTION_INITIATED) -> None:
    initiation_node = ScKeynodes.resolve(initiation, sc_type.CONST_NODE_CLASS)
    generate_connector(sc_type.CONST_PERM_POS_ARC, initiation_node, action_node)


def wait_agent(seconds: float, action_node: ScAddr, reaction_node: ScAddr = None) -> None:
    reaction_node = reaction_node or ScKeynodes[ActionStatus.ACTION_FINISHED]
    finish_event = Event()

    def event_callback(_: ScAddr, __: ScAddr, trg: ScAddr) -> ScResult:
        if trg != reaction_node:
            return ScResult.SKIP
        finish_event.set()
        return ScResult.OK

    event_params = ScEventSubscriptionParams(action_node, ScEventType.AFTER_GENERATE_INCOMING_ARC, event_callback)
    sc_event = create_elementary_event_subscriptions(event_params)[0]
    if not check_connector(sc_type.VAR_PERM_POS_ARC, reaction_node, action_node):
        finish_event.wait(seconds)
    destroy_elementary_event_subscriptions(sc_event)
    # TODO: return status in 0.2.0


def finish_action(action_node: ScAddr, status: Idtf = ActionStatus.ACTION_FINISHED) -> ScAddr:
    return generate_connector(sc_type.CONST_PERM_POS_ARC, ScKeynodes[status], action_node)


def finish_action_with_status(action_node: ScAddr, is_success: bool = True) -> None:
    status = ActionStatus.ACTION_FINISHED_SUCCESSFULLY if is_success else ActionStatus.ACTION_FINISHED_UNSUCCESSFULLY
    finish_action(action_node, status)
    finish_action(action_node, ActionStatus.ACTION_FINISHED)



"""
This source file is part of an OSTIS project. For the latest info, see https://github.com/ostis-ai
Distributed under the MIT License
(See an accompanying file LICENSE or a copy at https://opensource.org/licenses/MIT)
"""

import warnings
from typing import List, Optional, Union

from sc_client import client
from sc_client.constants import sc_type
from sc_client.constants.sc_type import ScType
from sc_client.models import ScAddr, ScConstruction, ScLinkContent, ScLinkContentType, ScTemplate, ScTemplateResult
from sc_client.models.sc_construction import ScLinkContentData

from sc_kpm.identifiers import CommonIdentifiers, ScAlias
from sc_kpm.sc_keynodes import Idtf, ScKeynodes


def generate_nodes(*node_types: ScType) -> List[ScAddr]:
    construction = ScConstruction()
    for node_type in node_types:
        construction.generate_node(node_type)
    return client.generate_elements(construction)


def create_nodes(*node_types: ScType) -> List[ScAddr]:
    warnings.warn(
        "Common utils 'create_nodes' method is deprecated. Use `generate_nodes` method instead.",
        DeprecationWarning,
    )
    return generate_nodes(*node_types)


def generate_node(node_type: ScType) -> ScAddr:
    return generate_nodes(node_type)[0]


def create_node(node_type: ScType) -> List[ScAddr]:
    warnings.warn(
        "Common utils 'create_node' method is deprecated. Use `generate_node` method instead.",
        DeprecationWarning,
    )
    return generate_nodes(node_type)


def generate_links(
    *contents: Union[str, int],
    content_type: ScLinkContentType = ScLinkContentType.STRING,
    link_type: ScType = sc_type.CONST_NODE_LINK,
) -> List[ScAddr]:
    construction = ScConstruction()
    for content in contents:
        link_content = ScLinkContent(content, content_type)
        construction.generate_link(link_type, link_content)
    return client.generate_elements(construction)


def create_links(
    *contents: Union[str, int],
    content_type: ScLinkContentType = ScLinkContentType.STRING,
    link_type: ScType = sc_type.CONST_NODE_LINK,
) -> List[ScAddr]:
    warnings.warn(
        "Common utils 'create_links' method is deprecated. Use `generate_links` method instead.",
        DeprecationWarning,
    )
    return generate_links(*contents, content_type=content_type, link_type=link_type)


def generate_link(
    content: Union[str, int],
    content_type: ScLinkContentType = ScLinkContentType.STRING,
    link_type: ScType = sc_type.CONST_NODE_LINK,
) -> ScAddr:
    return generate_links(content, content_type=content_type, link_type=link_type)[0]


def create_link(
    content: Union[str, int],
    content_type: ScLinkContentType = ScLinkContentType.STRING,
    link_type: ScType = sc_type.CONST_NODE_LINK,
) -> ScAddr:
    warnings.warn(
        "Common utils 'create_link' method is deprecated. Use `generate_link` method instead.",
        DeprecationWarning,
    )
    return generate_link(content, content_type=content_type, link_type=link_type)


def generate_connector(connector_type: ScType, src: ScAddr, trg: ScAddr) -> ScAddr:
    return generate_connectors(connector_type, src, trg)[0]


def create_edge(connector_type: ScType, src: ScAddr, trg: ScAddr) -> ScAddr:
    warnings.warn(
        "Common utils 'create_edge' method is deprecated. Use `generate_connector` method instead.",
        DeprecationWarning,
    )
    return generate_connector(connector_type, src, trg)


def generate_connectors(connector_type: ScType, src: ScAddr, *targets: ScAddr) -> List[ScAddr]:
    construction = ScConstruction()
    for trg in targets:
        construction.generate_connector(connector_type, src, trg)
    return client.generate_elements(construction)


def create_edges(connector_type: ScType, src: ScAddr, *targets: ScAddr) -> List[ScAddr]:
    warnings.warn(
        "Common utils 'create_edges' method is deprecated. Use `generate_connectors` method instead.",
        DeprecationWarning,
    )
    return generate_connectors(connector_type, src, *targets)


def generate_binary_relation(connector_type: ScType, src: ScAddr, trg: ScAddr, *relations: ScAddr) -> ScAddr:
    construction = ScConstruction()
    construction.generate_connector(connector_type, src, trg, ScAlias.RELATION_ARC)
    for relation in relations:
        construction.generate_connector(sc_type.CONST_PERM_POS_ARC, relation, ScAlias.RELATION_ARC)
    return client.generate_elements(construction)[0]


def create_binary_relation(connector_type: ScType, src: ScAddr, trg: ScAddr, *relations: ScAddr) -> ScAddr:
    warnings.warn(
        "Common utils 'create_binary_relation' method is deprecated. Use `generate_binary_relation` method instead.",
        DeprecationWarning,
    )
    return generate_binary_relation(connector_type, src, trg, *relations)


def generate_role_relation(src: ScAddr, trg: ScAddr, *rrel_nodes: ScAddr) -> ScAddr:
    return generate_binary_relation(sc_type.CONST_PERM_POS_ARC, src, trg, *rrel_nodes)


def create_role_relation(src: ScAddr, trg: ScAddr, *rrel_nodes: ScAddr) -> ScAddr:
    warnings.warn(
        "Common utils 'create_role_relation' method is deprecated. Use `generate_role_relation` method instead.",
        DeprecationWarning,
    )
    return generate_role_relation(src, trg, *rrel_nodes)


def generate_non_role_relation(src: ScAddr, trg: ScAddr, *nrel_nodes: ScAddr) -> ScAddr:
    return generate_binary_relation(sc_type.CONST_COMMON_ARC, src, trg, *nrel_nodes)


def create_norole_relation(src: ScAddr, trg: ScAddr, *nrel_nodes: ScAddr) -> ScAddr:
    warnings.warn(
        "Common utils 'create_norole_relation' method is deprecated. Use `generate_non_role_relation` method instead.",
        DeprecationWarning,
    )
    return generate_non_role_relation(src, trg, *nrel_nodes)


def check_connector(connector_type: ScType, source: ScAddr, target: ScAddr) -> bool:
    return bool(search_connectors(source, target, connector_type))


def check_edge(connector_type: ScType, source: ScAddr, target: ScAddr) -> bool:
    warnings.warn(
        "Common utils 'check_edge' method is deprecated. Use `check_connector` method instead.",
        DeprecationWarning,
    )
    return check_connector(connector_type, source, target)


def search_connector(source: ScAddr, target: ScAddr, connector_type: ScType) -> ScAddr:
    connectors = search_connectors(source, target, connector_type)
    return connectors[0] if connectors else ScAddr(0)


def get_edge(source: ScAddr, target: ScAddr, connector_type: ScType) -> ScAddr:
    warnings.warn(
        "Common utils 'get_edge' method is deprecated. Use `search_connector` method instead.",
        DeprecationWarning,
    )
    return search_connector(source, target, connector_type)


def search_connectors(source: ScAddr, target: ScAddr, *connector_types: ScType) -> List[ScAddr]:
    result_connectors = []
    for connector_type in connector_types:
        templ = ScTemplate()
        templ.triple(source, connector_type, target)
        results = client.search_by_template(templ)
        result_connectors.extend(result[1] for result in results)
    return result_connectors


def get_edges(source: ScAddr, target: ScAddr, *connector_types: ScType) -> List[ScAddr]:
    warnings.warn(
        "Common utils 'get_edges' method is deprecated. Use `search_connectors` method instead.",
        DeprecationWarning,
    )
    return search_connectors(source, target, *connector_types)


def get_element_system_identifier(addr: ScAddr) -> Idtf:
    nrel_system_idtf = ScKeynodes[CommonIdentifiers.NREL_SYSTEM_IDENTIFIER]

    templ = ScTemplate()
    templ.quintuple(
        addr,
        sc_type.VAR_COMMON_ARC,
        sc_type.VAR_NODE_LINK >> ScAlias.LINK,
        sc_type.VAR_PERM_POS_ARC,
        nrel_system_idtf,
    )
    result = client.search_by_template(templ)
    if result:
        return get_link_content_data(result[0].get(ScAlias.LINK))
    return ""


def get_system_idtf(addr: ScAddr) -> Idtf:
    warnings.warn(
        "Common utils 'get_system_idtf' method is deprecated. Use `get_element_system_identifier` method instead.",
        DeprecationWarning,
    )
    return get_element_system_identifier(addr)


def _search_relation_template(src: ScAddr, rel_node: ScAddr, rel_type: ScType) -> Optional[ScTemplateResult]:
    template = ScTemplate()
    template.quintuple(
        src,
        rel_type >> ScAlias.RELATION_ARC,
        sc_type.UNKNOWN >> ScAlias.ELEMENT,
        sc_type.VAR_PERM_POS_ARC,
        rel_node,
    )
    result = client.search_by_template(template)
    return result[0] if result else None


def search_role_relation_template(src: ScAddr, rrel_node: ScAddr) -> Optional[ScTemplateResult]:
    return _search_relation_template(src, rrel_node, sc_type.VAR_PERM_POS_ARC)


def search_non_role_relation_template(src: ScAddr, nrel_node: ScAddr) -> Optional[ScTemplateResult]:
    return _search_relation_template(src, nrel_node, sc_type.VAR_COMMON_ARC)


def search_norole_relation_template(src: ScAddr, nrel_node: ScAddr) -> Optional[ScTemplateResult]:
    warnings.warn(
        "Common utils 'search_norole_relation_template' method is deprecated."
        "Use `search_non_role_relation_template` method instead.",
        DeprecationWarning,
    )
    return search_non_role_relation_template(src, nrel_node)


def search_element_by_role_relation(src: ScAddr, rrel_node: ScAddr) -> ScAddr:
    search_result = search_role_relation_template(src, rrel_node)
    return search_result.get(ScAlias.ELEMENT) if search_result else ScAddr(0)


def get_element_by_role_relation(src: ScAddr, rrel_node: ScAddr) -> ScAddr:
    warnings.warn(
        "Common utils 'get_element_by_role_relation' method is deprecated."
        "Use `search_element_by_role_relation` method instead.",
        DeprecationWarning,
    )
    return search_element_by_role_relation(src, rrel_node)


def search_element_by_non_role_relation(src: ScAddr, nrel_node: ScAddr) -> ScAddr:
    search_result = search_non_role_relation_template(src, nrel_node)
    return search_result.get(ScAlias.ELEMENT) if search_result else ScAddr(0)


def get_element_by_norole_relation(src: ScAddr, nrel_node: ScAddr) -> ScAddr:
    warnings.warn(
        "Common utils 'get_element_by_norole_relation' method is deprecated."
        "Use `search_element_by_non_role_relation` method instead.",
        DeprecationWarning,
    )
    return search_element_by_non_role_relation(src, nrel_node)


def get_link_content_data(link: ScAddr) -> ScLinkContentData:
    content_part = client.get_link_content(link)
    return content_part[0].data


def erase_connectors(source: ScAddr, target: ScAddr, *connector_types: ScType) -> bool:
    return client.erase_elements(*search_connectors(source, target, *connector_types))


def delete_edges(source: ScAddr, target: ScAddr, *connector_types: ScType) -> bool:
    warnings.warn(
        "Common utils 'delete_edges' method is deprecated. Use `erase_connectors` method instead.",
        DeprecationWarning,
    )
    return erase_connectors(source, target, *connector_types)
